/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : Tue Dec 30 12:45:43 CST 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <SDL/SDL.h>
#include <SDL/SDL_net.h>

void Print_Usage();
int  Init();
void Quit();
int Tranceive(const char *host, Uint16 port,FILE *fp, int sending);

int main(int argc, char *argv[])
{
  Uint16 port=1234;

  if(Init()<0)
  {
    fprintf(stderr,"Couldn't initialize SDL/SDL_net!\n\n");
    return(2);
  }  

  atexit(Quit);
  
  if(argc<4)
  {
    Print_Usage();
    return(1);
  }

  port=atoi(argv[3]);

  if(strcmp(argv[1],"TO")==0)
  {
    if(argc==5)
    {
      FILE *read_from=fopen(argv[4],"r");
      if(read_from==NULL)
      {
        fprintf(stderr,"Couldn't open %s for read\n",argv[4]);
        return(5);
      }

      if(Tranceive(argv[2],port,read_from,1)<0)
        return(3);

      fclose(read_from);
    }
    else if(Tranceive(argv[2],port,stdin,1)<0)
      return(3);

    return(0);    
  }
  else if(strcmp(argv[1],"FROM")==0)
  {
    if(argc==5)
    {
      FILE *write_to=fopen(argv[4],"w");
      if(write_to==NULL)
      {
        fprintf(stderr,"Couldn't open %s for write\n",argv[4]);
        return(5);
      }
      else if(Tranceive(argv[2],port,write_to,0)<0)
        return(4);

      fclose(write_to);
    }
    else if(Tranceive(argv[2],port,stdout,0)<0)
      return(4);

    return(0);
  }
  else
  {
    Print_Usage();
    return(1);
  }

  return(0);  
}

int  Init()
{
  if(SDL_Init(0)<0)
    return(-1);
  else if(SDLNet_Init()<0)
  {
    SDL_Quit();
    return(-1);
  }
  else
    return(0);
}

void Quit()
{
  SDLNet_Quit();
  SDL_Quit();
}

int Tranceive(const char *host, Uint16 port, FILE *fp, int sending)
{
  int         running=1;  
  IPaddress        ip;
  TCPsocket      sock=NULL;
  const Uint8 *ipaddr=(Uint8 *)&(ip.host);

  if(SDLNet_ResolveHost(&ip,(char *)host,port)<0)
  {
    fprintf(stderr,"Couldn't resolve IP address from %s\n",host);
    return(-1);
  }
  else
    fprintf(stderr,"IP address resolved: %u.%u.%u.%u\n\n",
      ipaddr[0],ipaddr[1],ipaddr[2],ipaddr[3]             );
  
  
  if(sending)
  {
    Uint8  buf[512];

    signal(SIGPIPE,SIG_IGN);
    
    sock=SDLNet_TCP_Open(&ip);
    if(sock==NULL)
    {
      fprintf(stderr,"Couldn't open connection to %s:%u\n",host,port);
      return(-1);
    }
    fprintf(stderr,"Sending TO %s:%u\n",host,port);

    if(SDLNet_TCP_Recv(sock,buf,512)<=0)
    {
      fprintf(stderr,"Failed connection\n");
      SDLNet_TCP_Close(sock);
      return(-1);
    }
    else if(strncmp(buf,"OK",2)!=0)
    {
      fprintf(stderr,"Failed connection\n");
      SDLNet_TCP_Close(sock);
      return(-1);            
    }
    else
    {
      fprintf(stderr,"Recieved OK!\n");
    }

    while(running)
    {
      size_t bytes;


      if(feof(fp)||ferror(fp))
      {
        running=0;
        continue;
      }
      
      bytes=fread(buf,1,512,fp);
      if(bytes==0) continue;

      if(SDLNet_TCP_Send(sock,buf,bytes)<bytes)
      {
        fprintf(stderr,"Connection Broken!\n");
        SDLNet_TCP_Close(sock);
        sock=NULL;
        running=0;
        continue;
      }
    }

    fprintf(stderr,"Finished sending\n");
    
    if(sock==NULL) /* error - connection broken */
      return(-1);
    else
    {
      SDLNet_TCP_Close(sock);
      return(0);
    }
  }
  else /* Recieving */
  {
    TCPsocket hostsock=NULL;
    IPaddress hostip;
    
    if(SDLNet_ResolveHost(&hostip, NULL, port)<0)
    {
      fprintf(stderr,"Couldn't resolve listener socket on %u\n",port);
      return(-1);
    }
    
    hostsock=SDLNet_TCP_Open(&hostip);
    if(hostsock==NULL)
    {
      fprintf(stderr,"Couldn't create listener socket on %u\n",port);
      return(-1);
    }

    fprintf(stderr,"Waiting for connection from %s:%u...\n",host,port);

    while(running) /* Wait for proper connection */
    {      
      IPaddress *remote_ip=NULL;

      SDL_Delay(10);
      
      sock=SDLNet_TCP_Accept(hostsock);
      if(sock==NULL) continue;

      remote_ip=SDLNet_TCP_GetPeerAddress(sock);
      if(remote_ip==NULL)
      {
        SDLNet_TCP_Close(sock);
        sock=NULL;
        continue;
      }

      if(remote_ip->host==ip.host)
      {
        running=0;
        continue;
      }
      else
      {
        Uint8 *failip=(Uint8 *)&(remote_ip->host);
        fprintf(stderr,"Failed connection from %u.%u.%u.%u, ",
          failip[0],failip[1],failip[2],failip[3]);
        failip=(Uint8 *)&(ip.host);
        fprintf(stderr,"expected %u.%u.%u.%u\n",
          failip[0],failip[1],failip[2],failip[3]);

        SDLNet_TCP_Send(sock,"GO AWAY",strlen("GO AWAY"));
        remote_ip=NULL;
        SDLNet_TCP_Close(sock);
        sock=NULL;          
      }

    }

    SDLNet_TCP_Send(sock,"OK",strlen("OK"));

    /* Connection obtained! */
    
    fprintf(stderr,"!!! Recieving FROM %s:%u\n",host,port);

    running=1;
    while(running)
    {
      int bytes=0;
      char   buf[512];

      bytes=SDLNet_TCP_Recv(sock,buf,512);
      if(bytes<=0) /* end of transmission or connection broken */
      {
        running=0;
        continue;        
      }

      fwrite(buf,bytes,1,fp);
      if(ferror(fp))
      {
        fprintf(stderr,"Error in writing to file!\n");
        SDLNet_TCP_Close(sock);
        sock=NULL;
        running=0;
        continue;
      }
    }

    if(sock==NULL) /* Error in recieving */
    {
      SDLNet_TCP_Close(hostsock);
      hostsock=NULL;
      return(-1);
    }
    else /* success */
    {
      SDLNet_TCP_Close(sock);
      sock=NULL;
      SDLNet_TCP_Close(hostsock);
      hostsock=NULL;
      return(0);      
    }
  }
    
  return(0);
}

void Print_Usage()
{
  printf("USAGE:    tcp_pipe FROM this.host.ca port# [inpfile]\n"
         "          tcp_pipe TO 24.72.40.199 port# [inpfile]\n\n" );
}
